@@ -196,15 +196,42 @@ impl DataType {
         &self.parameters
     }
 
-    /// Returns the value of the given integer parameter.
-    pub fn int_parameter(&self, index: usize) -> Option<u64> {
-        if let Some(Parameter::Unsigned(value)) = self.parameters.get(index) {
+    /// Returns the value of the given boolean parameter.
+    pub fn bool_parameter(&self, index: usize) -> Option<bool> {
+        if let Some(Parameter::Boolean(value)) = self.parameters.get(index) {
             Some(*value)
         } else {
             None
         }
     }
 
+    /// Returns the value of the given integer parameter.
+    pub fn int_parameter(&self, index: usize) -> Option<i64> {
+        if let Some(Parameter::Integer(value)) = self.parameters.get(index) {
+            Some(*value)
+        } else {
+            None
+        }
+    }
+
+    /// Returns the value of the given enum parameter.
+    pub fn enum_parameter(&self, index: usize) -> Option<&str> {
+        if let Some(Parameter::Enum(value)) = self.parameters.get(index) {
+            Some(value)
+        } else {
+            None
+        }
+    }
+
+    /// Returns the value of the given string parameter.
+    pub fn string_parameter(&self, index: usize) -> Option<&str> {
+        if let Some(Parameter::String(value)) = self.parameters.get(index) {
+            Some(value)
+        } else {
+            None
+        }
+    }
+
     /// Returns the value of the given type parameter.
     pub fn type_parameter(&self, index: usize) -> Option<Arc<DataType>> {
         match self.parameters.get(index) {
@@ -462,13 +489,22 @@ pub trait ParameterInfo {
     /// Checks whether the given parameter set is valid for this base type.
     fn check_parameters(&self, params: &[Parameter]) -> diagnostic::Result<()>;
 
-    /// Returns the logical name of the given parameter.
+    /// Returns the logical name of the given parameter. If the index is out of
+    /// range or the name is unknown, yield None.
     fn parameter_name(&self, index: usize) -> Option<String>;
 
+    /// Returns the logical name of the given parameter, returning its index
+    /// if the name is not known.
+    fn parameter_name_or_index(&self, index: usize) -> String {
+        self.parameter_name(index)
+            .unwrap_or_else(|| format!("{}", index))
+    }
+
     /// Whether this type supports parameters. This is used to determine
     /// whether to print <> when the parameter list is empty. This is used to
     /// distinguish a concrete empty struct from a struct with unspecified
-    /// fields.
+    /// fields. If it's unknown whether a type has parameters, false is to be
+    /// returned.
     fn has_parameters(&self) -> bool;
 }
 
@@ -513,33 +549,24 @@ impl ParameterInfo for Class {
                 }
             }
             Class::Compound(compound) => compound.check_parameters(params),
-            Class::UserDefined(_) => {
-                if params.is_empty() {
-                    Ok(())
-                } else {
-                    Err(cause!(
-                        TypeMismatchedParameters,
-                        "user-defined types cannot currently be parameterized"
-                    ))
-                }
-            }
+            Class::UserDefined(user_defined) => user_defined.check_parameters(params),
             Class::Unresolved => Ok(()),
         }
     }
 
     fn parameter_name(&self, index: usize) -> Option<String> {
-        if let Class::Compound(compound) = self {
-            compound.parameter_name(index)
-        } else {
-            None
+        match self {
+            Class::Compound(compound) => compound.parameter_name(index),
+            Class::UserDefined(user_defined) => user_defined.parameter_name(index),
+            _ => None,
         }
     }
 
     fn has_parameters(&self) -> bool {
-        if let Class::Compound(compound) = self {
-            compound.has_parameters()
-        } else {
-            false
+        match self {
+            Class::Compound(compound) => compound.has_parameters(),
+            Class::UserDefined(user_defined) => user_defined.has_parameters(),
+            _ => false,
         }
     }
 }
@@ -602,13 +629,13 @@ impl ParameterInfo for Compound {
                         "{self} expects a single parameter (length)"
                     ));
                 }
-                if let Parameter::Unsigned(length) = params[0] {
+                if let Parameter::Integer(length) = params[0] {
                     // Note: 2147483647 = 2^31-1 = maximum value for signed
                     // 32-bit integer. However, the significance of the number
                     // is just that the Substrait specification says this is
                     // the limit.
-                    const MIN_LENGTH: u64 = 1;
-                    const MAX_LENGTH: u64 = 2147483647;
+                    const MIN_LENGTH: i64 = 1;
+                    const MAX_LENGTH: i64 = 2147483647;
                     if !(MIN_LENGTH..=MAX_LENGTH).contains(&length) {
                         return Err(cause!(
                             TypeMismatchedParameters,
@@ -629,17 +656,17 @@ impl ParameterInfo for Compound {
                         "{self} expects two parameters (precision and scale)"
                     ));
                 }
-                if let Parameter::Unsigned(precision) = params[0] {
-                    const MIN_PRECISION: u64 = 1;
-                    const MAX_PRECISION: u64 = 38;
+                if let Parameter::Integer(precision) = params[0] {
+                    const MIN_PRECISION: i64 = 1;
+                    const MAX_PRECISION: i64 = 38;
                     if !(MIN_PRECISION..=MAX_PRECISION).contains(&precision) {
                         return Err(cause!(
                             TypeMismatchedParameters,
                             "{self} precision {precision} is out of range {MIN_PRECISION}..{MAX_PRECISION}"
                         ));
                     }
-                    if let Parameter::Unsigned(scale) = params[1] {
-                        if scale > precision {
+                    if let Parameter::Integer(scale) = params[1] {
+                        if scale < 0 || scale > precision {
                             return Err(cause!(
                                 TypeMismatchedParameters,
                                 "{self} scale {scale} is out of range 0..{precision}"
@@ -745,17 +772,202 @@ impl ParameterInfo for Compound {
     }
 }
 
+impl ParameterInfo for extension::Reference<extension::DataType> {
+    fn check_parameters(&self, params: &[Parameter]) -> diagnostic::Result<()> {
+        self.definition
+            .as_ref()
+            .map(|d| d.check_parameters(params))
+            .unwrap_or(Ok(()))
+    }
+
+    fn parameter_name(&self, index: usize) -> Option<String> {
+        self.definition
+            .as_ref()
+            .map(|d| d.parameter_name(index))
+            .unwrap_or_default()
+    }
+
+    fn has_parameters(&self) -> bool {
+        self.definition
+            .as_ref()
+            .map(|d| d.has_parameters())
+            .unwrap_or_default()
+    }
+}
+
+impl ParameterInfo for extension::DataType {
+    fn check_parameters(&self, params: &[Parameter]) -> diagnostic::Result<()> {
+        // Determine the minimum number of parameters and check whether we have
+        // enough.
+        let min_parameters = self
+            .parameter_slots
+            .iter()
+            .enumerate()
+            .rev()
+            .find_map(|(index, slot)| if slot.optional { None } else { Some(index + 1) })
+            .unwrap_or_default();
+        if params.len() < min_parameters {
+            return Err(cause!(
+                TypeMismatchedParameters,
+                "need at least {min_parameters} parameter(s), but only {} was/were provided",
+                params.len()
+            ));
+        }
+
+        // Match parameters to slots positionally.
+        for (index, param) in params.iter().enumerate() {
+            // Determine the slot that corresponds to this parameter.
+            let slot = self
+                .parameter_slots
+                .get(index)
+                .or_else(|| {
+                    if self.parameters_variadic {
+                        self.parameter_slots.last()
+                    } else {
+                        None
+                    }
+                })
+                .ok_or_else(|| {
+                    cause!(
+                        TypeMismatchedParameters,
+                        "type expects at most {index} parameters, but {} were provided",
+                        params.len()
+                    )
+                })?;
+
+            // Check the provided parameter against the information contained
+            // in the slot.
+            match param {
+                Parameter::Unresolved => (),
+                Parameter::Null => {
+                    if !slot.optional {
+                        return Err(cause!(
+                            TypeMismatchedParameters,
+                            "parameter {} is not optional and can thus not be skipped with null",
+                            self.parameter_name_or_index(index)
+                        ));
+                    }
+                }
+                Parameter::Boolean(_) => {
+                    if !matches!(slot.bounds, extension::DataTypeParameterBounds::Boolean) {
+                        return Err(cause!(
+                            TypeMismatchedParameters,
+                            "parameter {} must be bound to a boolean",
+                            self.parameter_name_or_index(index)
+                        ));
+                    }
+                }
+                Parameter::Integer(value) => {
+                    if let extension::DataTypeParameterBounds::Integer(min, max) = &slot.bounds {
+                        if value < min || value > max {
+                            return Err(cause!(
+                                TypeMismatchedParameters,
+                                "parameter {} is out of range {min}..{max}",
+                                self.parameter_name_or_index(index)
+                            ));
+                        }
+                    } else {
+                        return Err(cause!(
+                            TypeMismatchedParameters,
+                            "parameter {} must be bound to an integer",
+                            self.parameter_name_or_index(index)
+                        ));
+                    }
+                }
+                Parameter::Enum(value) => {
+                    if let extension::DataTypeParameterBounds::Enum(options) = &slot.bounds {
+                        if !options
+                            .iter()
+                            .any(|option| option.eq_ignore_ascii_case(value))
+                        {
+                            return Err(cause!(
+                                TypeMismatchedParameters,
+                                "parameter {} must be set to one of {:?}",
+                                self.parameter_name_or_index(index),
+                                options
+                            ));
+                        }
+                    } else {
+                        return Err(cause!(
+                            TypeMismatchedParameters,
+                            "parameter {} must be bound to an enum option",
+                            self.parameter_name_or_index(index)
+                        ));
+                    }
+                }
+                Parameter::String(_) => {
+                    if !matches!(slot.bounds, extension::DataTypeParameterBounds::String) {
+                        return Err(cause!(
+                            TypeMismatchedParameters,
+                            "parameter {} must be bound to a string",
+                            self.parameter_name_or_index(index)
+                        ));
+                    }
+                }
+                Parameter::Type(_) | Parameter::NamedType(_, _) => {
+                    if !matches!(slot.bounds, extension::DataTypeParameterBounds::DataType) {
+                        return Err(cause!(
+                            TypeMismatchedParameters,
+                            "parameter {} must be bound to a data type",
+                            self.parameter_name_or_index(index)
+                        ));
+                    }
+                }
+            }
+        }
+        Ok(())
+    }
+
+    fn parameter_name(&self, index: usize) -> Option<String> {
+        if self.parameters_variadic && index + 1 >= self.parameter_slots.len() {
+            if let Some(slot) = self.parameter_slots.last() {
+                return Some(if slot.name.is_empty() {
+                    format!("{}", index)
+                } else {
+                    format!("{}.{}", slot.name, index + 1 - self.parameter_slots.len())
+                });
+            }
+        }
+        self.parameter_slots.get(index).map(|slot| {
+            if slot.name.is_empty() {
+                format!("{}", index)
+            } else {
+                slot.name.clone()
+            }
+        })
+    }
+
+    fn has_parameters(&self) -> bool {
+        !self.parameter_slots.is_empty()
+    }
+}
+
 /// Parameter for parameterized types.
 #[derive(Clone, Debug, PartialEq)]
 pub enum Parameter {
+    /// Unresolved. Used when the parameter could not be parsed.
+    Unresolved,
+
+    /// Null, to skip optional parameters.
+    Null,
+
+    /// Boolean type parameter (only used by extensions).
+    Boolean(bool),
+
+    /// Integer type parameter (varchar length, etc.).
+    Integer(i64),
+
+    /// Enumeration type parameter (only used by extensions).
+    Enum(String),
+
+    /// String type parameter (only used by extensions).
+    String(String),
+
     /// Type parameter (list element type, struct element types, etc).
     Type(Arc<DataType>),
 
     /// Named type parameter (named struct/schema pseudotype elements).
     NamedType(String, Arc<DataType>),
-
-    /// Integral type parameter (varchar length, etc.).
-    Unsigned(u64),
 }
 
 impl Describe for Parameter {
@@ -765,6 +977,12 @@ impl Describe for Parameter {
         limit: util::string::Limit,
     ) -> std::fmt::Result {
         match self {
+            Parameter::Unresolved => write!(f, "!"),
+            Parameter::Null => write!(f, "null"),
+            Parameter::Boolean(value) => write!(f, "{value}"),
+            Parameter::Integer(value) => write!(f, "{value}"),
+            Parameter::Enum(variant) => util::string::describe_identifier(f, variant, limit),
+            Parameter::String(value) => util::string::describe_string(f, value, limit),
             Parameter::Type(data_type) => data_type.describe(f, limit),
             Parameter::NamedType(name, data_type) => {
                 let (name_limit, type_limit) = limit.split(name.len());
@@ -772,11 +990,16 @@ impl Describe for Parameter {
                 write!(f, ": ")?;
                 data_type.describe(f, type_limit)
             }
-            Parameter::Unsigned(value) => write!(f, "{value}"),
         }
     }
 }
 
+impl Default for Parameter {
+    fn default() -> Self {
+        Parameter::Unresolved
+    }
+}
+
 impl std::fmt::Display for Parameter {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         self.display().fmt(f)
@@ -857,8 +1080,20 @@ impl From<Arc<DataType>> for Parameter {
     }
 }
 
-impl From<u64> for Parameter {
-    fn from(x: u64) -> Self {
-        Parameter::Unsigned(x)
+impl From<i64> for Parameter {
+    fn from(x: i64) -> Self {
+        Parameter::Integer(x)
+    }
+}
+
+impl From<bool> for Parameter {
+    fn from(x: bool) -> Self {
+        Parameter::Boolean(x)
+    }
+}
+
+impl From<String> for Parameter {
+    fn from(x: String) -> Self {
+        Parameter::String(x)
     }
 }
