@@ -188,7 +188,7 @@ impl Describe for Literal {
                         if d < &0 {
                             write!(f, "-")?;
                         }
-                        let d = d.abs() as u128;
+                        let d = d.unsigned_abs();
                         let s = 10u128.pow(scale as u32);
                         if self
                             .data_type
@@ -670,7 +670,7 @@ fn parse_fixed_char(
         data_type::Compound::FixedChar,
         nullable,
         variation,
-        vec![x.len() as u64],
+        vec![i64::try_from(x.len()).unwrap()],
     )
 }
 
@@ -682,9 +682,9 @@ fn parse_var_char(
     variation: Option<Arc<extension::Reference<extension::TypeVariation>>>,
 ) -> diagnostic::Result<Literal> {
     proto_primitive_field!(x, y, length);
-    let len = x.length as usize;
+    let len: u32 = x.length;
     proto_primitive_field!(x, y, value, |x, _| {
-        if x.len() > len {
+        if x.len() > len as usize {
             Err(cause!(
                 ExpressionIllegalLiteralValue,
                 "varchar literal value is longer than specified length"
@@ -698,7 +698,7 @@ fn parse_var_char(
         data_type::Compound::VarChar,
         nullable,
         variation,
-        vec![len as u64],
+        vec![len as i64],
     )
 }
 
@@ -714,7 +714,7 @@ fn parse_fixed_binary(
         data_type::Compound::FixedBinary,
         nullable,
         variation,
-        vec![x.len() as u64],
+        vec![i64::try_from(x.len()).unwrap()],
     )
 }
 
@@ -748,8 +748,8 @@ fn parse_decimal(
         }
     })
     .1;
-    let precision = u64::try_from(x.precision).unwrap_or_default();
-    let scale = u64::try_from(x.scale).unwrap_or_default();
+    let precision = x.precision as i64;
+    let scale = x.scale as i64;
 
     if let Some(val) = val {
         let range = 10i128.saturating_pow(precision.try_into().unwrap_or_default());
